from .command import Command
from ..dotfile import ConfigItem
from pathlib import Path
from ..utils import link, create_traceback
import doter.events as events
import asyncio

__CLASS_NAME__ = "Link"
__COMMAND_NAME__ = "link"


class Link(Command):
    def __init__(self, config, dotfiles_dir, event_bus):
        super().__init__(config, dotfiles_dir, event_bus)

    async def __call__(self, *args, force=False):
        try:
            config_items = args if len(args) > 0 else self._config.all_items
            print(self._config)
            await asyncio.gather(*[
                self._do_link(self._config.items[key], force)
                for key in config_items
            ])
        except KeyError as e:
            traceback = create_traceback(KeyError, e, e.__traceback__)
            await self._event_bus.publish(
                events.TASK_ERROR,
                name=None,
                description=f'Key does not exist: {str(e)}',
                traceback=traceback,
            )

    async def _do_link(self, config: ConfigItem, force: bool):
        try:
            for src, dst in config.files.items():
                abs_src = Path(src)
                if (not force) and abs_src.exists():
                    raise RuntimeError(
                        f'{src} already exists, and "--force is not set"')
                link(self._dotfiles_dir, src, dst, force)
        except RuntimeError as e:
            await self._publish(
                events.ERROR,
                descrption=str(e),
                traceback=e,
            )



from doter import events
from doter.dotfile import ConfigItem
from .command import Command
from ..utils import create_traceback
import asyncio
from ..utils import link, run_hooks
from pathlib import Path

__CLASS_NAME__ = "Install"
__COMMAND_NAME__ = "install"


class Install(Command):
    def __init__(self, dotfiles, dotfiles_dir, event_bus):
        super().__init__(dotfiles, dotfiles_dir, event_bus)

    async def __call__(self, *args, force=False):
        try:
            config_items = args if len(args) > 0 else self._config.all_items
            print(self._config)
            await asyncio.gather(*[
                self._do_isntall(key, self._config.items[key], force)
                for key in config_items
            ])
        except KeyError as e:
            traceback = create_traceback(KeyError, e, e.__traceback__)
            await self._event_bus.publish(
                events.TASK_ERROR, 
                name=None,
                description=f'Key does not exist: {str(e)}',
                traceback=traceback,
            )
            return

    async def _do_isntall(self, key: str, config: ConfigItem, force: bool):
        pre_install_hooks = config.pre_hooks
        post_install_hooks = config.post_hooks
        total_steps = len(pre_install_hooks) + len(post_install_hooks) + len(config.files)
        await self._event_bus.publish(
            events.TASK_INIT,
            name=key,
            description='Preparing for setup',
            total=total_steps,
        )
        try:
            # Run pre-install hoooks
            await run_hooks(pre_install_hooks, self._event_bus, key)
        except RuntimeError as e:
            traceback = create_traceback(RuntimeError, e, e.__traceback__)
            await self._event_bus.publish(
                events.TASK_ERROR,
                name=key,
                description=f'Pre-Install hook: {str(e)}',
                traceback=traceback,
            )
            return

        # make symlink
        try:
            for src, dst in config.files.items():
                abs_src = Path(src)
                if (not force) and abs_src.exists():
                    raise RuntimeError(
                        f'{src} already exists, and "--force is not set"')
                link(self._dotfiles_dir, src, dst, force)
        except RuntimeError as e:
            await self._publish(
                events.TASK_ERROR,
                name=key,
                descrption=str(e),
                traceback=e,
            )

        try:
            # run post-install hooks
            await run_hooks(post_install_hooks, self._event_bus, key)
        except RuntimeError as e:
            traceback = create_traceback(RuntimeError, e, e.__traceback__)
            await self._event_bus.publish(
                events.TASK_ERROR,
                name=key,
                traceback=traceback,
            )
            return
        await self._event_bus.publish(events.TASK_DONE, name=key)




from typing import List, Union, Type
from types import TracebackType
from shutil import get_terminal_size
import os
import asyncio
from pathlib import Path
from rich.traceback import Traceback
import doter.events as events
import signal


async def run_hooks(hooks: List[str], event_bus=None, name=None):
    """
    run_hooks execute hooks through the system shell
    asyncriously

    Args:
        hooks (List[str]): a series of commands to be executed

    Raises:
        RuntimeError: execution will be terminated when a command returns
            non zero
    """
    home_dir = os.getenv('HOME')
    if home_dir is None:
        raise RuntimeError('environment variable $HOME is not set')
    for cmd in hooks:
        if name is not None and event_bus is not None:
            await event_bus.publish(
                events.TASK_UPDATE,
                name=name,
                description=f"Executing hook: {cmd}",
            )
        proc = await asyncio.create_subprocess_shell(
            cmd.replace('~', home_dir),
            stdout=asyncio.subprocess.DEVNULL,
            stderr=asyncio.subprocess.PIPE,
            shell=True)

        async def sigint_handler():
            proc.send_signal(signal.SIGINT)
            await proc.wait()

        event_bus.subscribe('install/sigint', sigint_handler)
        _, stderr = await proc.communicate()
        retcode = proc.returncode
        if retcode != 0:
            raise RuntimeError(
                f'Hook {cmd} exited with code {retcode}: {stderr.decode()}')


def link(dotfiles_dir: Union[str, Path], src: str, dst: str, force: bool):
    """
    link creates a symbolic link from dst to src
    dst are the files in dotfiles_dir

    Args:
        dotfiles_dir (str): the directory containing dotfiles
        src (str): the source file
        dst (str): the destination file
        force (bool): whether to overwrite existing files
    """
    abs_src = os.path.abspath(os.path.expanduser(src))
    abs_dst = os.path.join(dotfiles_dir, os.path.expanduser(dst))
    abs_dst = os.path.abspath(abs_dst)
    abs_src = Path(abs_src)
    if force and abs_src.is_symlink():
        abs_src.unlink()
        print(f"Unlinking {abs_src}")
    elif force and abs_src.is_file():
        os.remove(abs_src)
        print(f"Removing {abs_src}")
    elif force and abs_src.is_dir():
        os.removedirs(abs_src)
    abs_src.parent.mkdir(parents=True, exist_ok=True)
    os.symlink(src=abs_dst, dst=abs_src)


async def sh(cmd: str):
    """
    sh executes a shell command
    """
    proc = await asyncio.create_subprocess_shell(
        cmd,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        shell=True)
    stdout, stderr = await proc.communicate()
    retcode = proc.returncode
    return retcode, stdout, stderr


def create_traceback(exc_type: Type[BaseException], exec_val: BaseException,
                     traceback: Union[TracebackType, None]):
    """
    create_traceback creates an traceback object in `rich`
    from an exception

    Args:
        exec_type (Type[BaseException]): Type of the Exception
        exec_val (BaseException): The execption instance
        traceback(TracebackType): Optional, the traceback object

    Returns
        Traceback (rich.traceback.Traceback): the printable traceback

    """
    width, _ = get_terminal_size((80, 20))
    return Traceback.from_exception(
        exc_type,
        exec_val,
        traceback,
        width=width,
        word_wrap=True,
        show_locals=True,
    )



from typing import TypedDict, List, Dict


class ConfigItem(TypedDict, total=False):
    src: str
    dst: str
    before_setup: List[str]
    after_steup: List[str]


class GlobalHooks(TypedDict, total=False):
    before_setup: List[str]
    after_setup: List[str]


Dotfiles = Dict[str, ConfigItem]


class ConfigYaml(TypedDict, total=False):
    hooks: GlobalHooks
    files: Dotfiles



from pathlib import Path
import signal
import asyncio
import importlib
import os
from rich.traceback import install as install_traceback

from typing import Any
from .eventbus import EventBus
from .ui import RichUI

EXCLUDED_FILES = [
    '.git', '.gitignore', '.gitmodules', '.DS_Store', '__pycache__',
    'command.py'
]


class DoterApp(object):
    """  class for the app
    DoterApp contains for all commands for Fire to look up,
    it will source the commands from the commands modules.
    It will also contain the UI component for the app.

    Note: For Details on the commands see the commands modules
    
    Args:
        config (str): path to the config file
        dotfiles_dir (str): path to the dotfiles directory

    """
    _instance = None

    def __init__(self,
                 config='~/.dotfiles/config.yml',
                 dotfiles_dir='~/.dotfiles'):
        install_traceback(show_locals=True)

        self._config_path = Path(config).expanduser().absolute()
        self._dotfiles_dir = Path(dotfiles_dir).expanduser().absolute()
        self._event_bus = EventBus()
        self._ui_component = RichUI(self._event_bus)
        self._resolve_modules()
        DoterApp._instance = self

    def _resolve_modules(self):
        """
        Resolve the commands modules
        it will import every python source file that is not in the
        "EXCLUDED_FILES". If `__CLASS_NAME__` and `__COMMAND_NAME__`
        it will instantiate the class and bind it to the given
        `__COMMAND_NAME__`

    
        """
        commands_dir = Path(__file__).parent / 'commands'
        for fn in os.listdir(commands_dir):
            if fn.endswith('.py') and not fn.startswith(
                    '_') and fn not in EXCLUDED_FILES:
                module: Any = importlib.import_module(
                    'doter.commands.{}'.format(fn[:-3]))
                class_name = module.__CLASS_NAME__
                cmd_name = module.__COMMAND_NAME__
                cls = getattr(module, class_name)
                cmd = cls(self._config_path, self._dotfiles_dir,
                          self._event_bus)
                wrapped_cmd = self._ui_wrapper(cmd.__call__)
                setattr(self, cmd_name, wrapped_cmd)

    def _ui_wrapper(self, func):
        """
        _ui_wrapper is used to wrap the command function
        so the progress bar can be started and ended properly

        Args:
            func (callable): the command function / method to be wrapped

        Returns:
            callable: the wrapped function
        """
        async def inner(*args, **kwargs):
            try:
                self._ui_component.progress.start()
                await func(*args, **kwargs)
            except:
                self._ui_component.progress.console.print_exception(
                    max_frames=20)
                self._ui_component.progress.stop()
                exit(1)
            finally:
                self._ui_component.progress.stop()

        return inner

    @classmethod
    def _get_instance(cls):
        """
        _get_instance is used to get current instance exteranlly,
        primarily used in handle Exceptions
        
        Returns:
            DoterApp 
        """
        if cls._instance is None:
            raise ValueError("No running DoterApp instance")
        return cls._instance

    def _stop(self):
        """
        _stop stops the progress bar without using the eventbus

        NOTE: this method should only be used when handling exceptions
        """
        self._ui_component.progress.stop()


if __name__ == '__main__':
    app = DoterApp()



from dataclasses import dataclass
from typing import Dict, List, Union
import yaml


@dataclass
class ConfigItem:
    files: Dict[str, str]
    pre_hooks: List[str]
    post_hooks: List[str]
    clean_hooks: Union[List[str], None]


@dataclass
class Dotfile:
    items: Dict[str, ConfigItem]
    all_items: List[str]
    _raw: dict


def parse_config(path: str):
    f = open(path, 'r')
    dotfile = yaml.load(f, Loader=yaml.Loader)
    all_items = dotfile.get('all', None)
    if all_items is None:
        all_items = list(dotfile['config'].keys())
    items = {}
    f.close()
    for key, item in dotfile['config'].items():
        config_item = ConfigItem(files=item.get('files', dict()),
                                 pre_hooks=item.get('before_setup', []),
                                 post_hooks=item.get('after_setup', []),
                                 clean_hooks=item.get('clean_hooks', []))
        items[key] = config_item
    return Dotfile(items=items, all_items=all_items, _raw=dotfile)



from .command import Command
import asyncio
from ..utils import create_traceback
from ..dotfile import ConfigItem
from pathlib import Path
import doter.events as events

__CLASS_NAME__ = "Clean"
__COMMAND_NAME__ = "clean"


class Clean(Command):
    def __init__(self, dotfiles, dotfiles_dir, event_bus):
        super().__init__(dotfiles, dotfiles_dir, event_bus)

    async def __call__(self, *args):
        try:
            config_items = args if len(args) > 0 else self._config.all_items
            print(self._config)
            await asyncio.gather(*[
                self._do_clean(self._config.items[key]) for key in config_items
            ])
        except KeyError as e:
            traceback = create_traceback(KeyError, e, e.__traceback__)
            await self._event_bus.publish(
                events.TASK_ERROR,
                name=None,
                description=f'Key does not exist: {str(e)}',
                traceback=traceback,
            )

    async def _do_clean(self, item: ConfigItem):
        try:
            for src in item.files.keys():
                src_path = Path(src).expanduser().absolute()
                if src_path.is_symlink():
                    src_path.unlink()
                    await self._publish(events.DONE, message=f'removed {src}')
                else:
                    raise RuntimeWarning(
                        f'Skipping {src}: not a symbolic link')
        except RuntimeWarning as warning:
            await self._publish(events.WARN, message=str(warning))



from typing import List, Callable, Dict

from dataclasses import dataclass
import asyncio


@dataclass
class EventBus(object):
    """
    EventBus is used to publish and dispatch events in
    an async manner. Currently used for UI
    """
    _subscribers: Dict[str, List[Callable]]

    def __init__(self):
        self._subscribers = dict()

    def subscribe(self, topic: str, callback: Callable):
        """
        Subscribe to an event with a callback function function

        Args:
            callback (Callable): The function to be called when
            the event is published. It should take keyword arguments (**kwargs)
        """
        if topic not in self._subscribers.keys():
            self._subscribers[topic] = [callback]
            return
        self._subscribers[topic].append(callback)

    async def publish(self, topic: str, **kwargs):
        """
        publish an event asyncrously

        Args:
            topic (str): the identifier of the event
            **kwrags: the extra context to be passed to the subscribers
        """
        subscribers = self._subscribers.get(topic, None)
        if subscribers != None:
            return await asyncio.gather(
                *[asyncio.create_task(s(**kwargs)) for s in subscribers], )
        return None



from pathlib import Path
from ..eventbus import EventBus
from ..dotfile import parse_config


class Command:
    def __init__(self, config_path, dotfiles_dir, event_bus: EventBus):
        self._config = parse_config(config_path)
        self._config_path = config_path
        self._event_bus = event_bus
        self._dotfiles_dir = Path(dotfiles_dir)

    def __call__(self):
        raise NotImplementedError()

    @property
    def _dotfiles(self):
        return self._config.items

    async def _publish(self, topic, **kwargs):
        await self._event_bus.publish(topic, **kwargs)



from typing import List, Tuple, Union

from doter import events
from .command import Command
from pathlib import Path
from shutil import move
from ..utils import create_traceback, link

import yaml

__CLASS_NAME__ = "Backup"
__COMMAND_NAME__ = "backup"

Args = Union[List[str], Tuple[str]]


class Backup(Command):
    def __init__(self, dotfiles, dotfiles_dir, event_bus):
        super().__init__(dotfiles, dotfiles_dir, event_bus)

    async def __call__(self, *args):
        if len(args) < 2:
            await self._event_bus.publish(
                events.ERROR, message='Usage: <name> <dotfile>:<target>')
            return
        name, specs = self.parse_args(args)
        try:
            for src, dst in specs:
                src_path = Path(src).expanduser().absolute()
                dst_path = Path(dst).expanduser().absolute()
                if not src_path.exists():
                    raise FileNotFoundError(src)

                move(src_path, dst_path)
                link(self._dotfiles_dir, str(src_path), str(dst_path), True)

            with open(self._config_path, 'w') as f:
                conf = self._config._raw
                files = dict(specs)
                if name not in conf['config'].keys():
                    conf['config'][name] = {'files': files}
                else:
                    conf['config'][name]['files'] = files
                yaml.dump(conf, f)
            await self._event_bus.publish('backup/done')
        except FileNotFoundError as err:
            tb = create_traceback(FileNotFoundError, err, err.__traceback__)
            await self._event_bus.publish(events.ERROR,
                                          message="dotfile not found",
                                          traceback=tb)

    def parse_args(self, args: Args) -> Tuple[str, List[Tuple[str, str]]]:
        task = args[0]
        res = []
        for arg in args[1::]:
            items = arg.split(':')
            if len(items) != 2:
                raise ValueError(f'Invalid spec for backup ({arg})')
            res.append((items[0], items[1]))

        return task, res



from rich.progress import Progress, TextColumn, SpinnerColumn
from rich.console import Console
from .eventbus import EventBus

import doter.events as events


class RichUI(object):
    """ RichUI

    A fancy command line progress display, it listens to
    events from the event bus

    """
    STATUS_EMOJIS = {
        "completed": ":white_check_mark:",
        "error": ":cross_mark:",
        "running": ":construction:",
        "warning": ":yellow_circle:"
    }

    def __init__(self, event_bus: EventBus):
        """
        Args:
            event_bus (EventBus): the event bus to be subscribed

        Attributes:
            _progress: (Rich.progress.Progress): the progress display
            _task_dict: the mapping of task name to task_id

        
        """
        self._task_dict = dict()
        self._progress = Progress(
            TextColumn("{task.fields[status]}"),  # symbol Column
            SpinnerColumn(),
            TextColumn("{task.fields[name]}"),  # task name column
            TextColumn("({task.completed}/{task.total})"),
            TextColumn("{task.description}"))
        self._console = Console(record=True)
        event_bus.subscribe(events.TASK_INIT, self._on_install_init)
        event_bus.subscribe(events.TASK_UPDATE, self._on_install_update)
        event_bus.subscribe(events.TASK_DONE, self._on_install_completed)
        event_bus.subscribe(events.TASK_ERROR, self._on_install_error)
        event_bus.subscribe(events.DONE, self._on_complete)
        event_bus.subscribe(events.ERROR, self._on_error)

    async def _on_install_error(self, **kwargs):
        """
        Callback function when an error occured
        """
        # extract information from context
        name = kwargs['name']
        task_id = self._task_dict.get(name, None)
        tb = kwargs['traceback']

        # Update progress to display error
        # and display traceback
        if task_id:
            self._progress.update(
                task_id,
                description="Failed",
                status=self.STATUS_EMOJIS['error'],
            )
        self._progress.console.print(tb)

    async def _on_install_completed(self, **params):
        """
        Callback when a task is completed
        """
        # extract variable from context
        name = params['name']
        description = f'Installation of {name} completed'
        task_id = self._task_dict[name]
        # Update the progress bar to display completion message
        self._progress.update(
            task_id,
            description=description,
            status=self.STATUS_EMOJIS['completed'],
            advance=1,
        )

    async def _on_install_init(self, **params):
        """
        Callback for a task initialised
        """
        # extract info
        name = params['name']
        description = params['description']
        total = params['total']

        # Update the progress bar
        task_id = self._progress.add_task(
            description,
            total=total,
            name=name,
            status=self.STATUS_EMOJIS['running'],
        )
        self._task_dict[name] = task_id

    async def _on_install_update(self, **params):
        name = params['name']
        description = params['description']
        task_id = self._task_dict[name]
        self._progress.update(
            task_id,
            advance=1,
            description=description,
        )

    async def on_sigint(self):
        for name, task_id in self._task_dict.items():
            self._progress.update(
                task_id,
                name=name,
                description="Cancelled",
                status=self.STATUS_EMOJIS['error'],
            )

    async def _on_complete(self, **kwargs):
        self._console.print(self.STATUS_EMOJIS['completed'] + " " +
                            kwargs['message'])

    async def _on_error(self, **kwargs):
        tb = kwargs.get('traceback', None)
        if tb:
            self._console.print(tb)
        self._console.print(
            f"{self.STATUS_EMOJIS['error']} {kwargs['message']}")

    async def _on_warning(self, **kwargs):
        tb = kwargs.get('traceback', None)
        if tb:
            self._console.print(tb)
        self._console.print(
            f"{self.STATUS_EMOJIS['warning']} {kwargs['message']}")

    @property
    def progress(self):
        return self._progress



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['doter', 'doter.commands']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0', 'fire>=0.4.0,<0.5.0', 'rich>=10.15.2,<11.0.0']

entry_points = \
{'console_scripts': ['doter = doter.__main__:main']}

setup_kwargs = {
    'name': 'doter',
    'version': '1.0.1',
    'description': 'A dotfile manager',
    'long_description': None,
    'author': 'Alex Zhang',
    'author_email': 'zhangchi0104@live.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
